#!/usr/bin/env python

##############################################################################
#                                                                            #
#    mail2news - A program to convert email to news articles and post them   #
#                via the NNTP POST command to a server                       #
#                                                                            #
#    Written by Joerg Dietrich <joerg@dietrich.net> and Copyright 2004.       #
#                                                                            #
#                                                                            #
#    This program is free software; you can redistribute it and/or modify    #
#    it under the terms of the GNU General Public License version 2 as       #
#    published by the Free Software Foundation.                              #
#                                                                            #
#    This program is distributed in the hope that it will be useful,         #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of          #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
#    GNU General Public License for more details.                            #
#                                                                            #
#    You should have received a copy of the GNU General Public License       #
#    along with this program; if not, write to the Free Software             #
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               #
#                                                                            #
##############################################################################

import os
import pwd
import re
import socket
import string
import sys
import tempfile
import traceback

import nntplib
import StringIO
import email
from email.Message import Message
from email.Header import Header

M2NVersion = "0.0.3"

# Flags to indicate what to do with headers
M2Nrm    = 0x01 # Remove this header from mail
M2Njoin  = 0x02 # Join fields if Header is duplicated
M2Nonce  = 0x04 # Header may appear not more than once
M2Nconv  = 0x08 # Convert header to news format
M2Nmatch = 0x10 # Use header to match regExp patterns
M2Nmust  = 0x20 # Header must exist in news article

headDict = {"cc"               : M2Nmatch|M2Njoin,
            "date"             : M2Nonce,
            "date-received"    : M2Nrm,
            "delivered-to"     : M2Nmatch,
            "folllowup-to"     : M2Nrm,
            "from"             : M2Nmust|M2Nmatch|M2Nonce,
            "in-reply-to"      : M2Nconv,
            "list-id"          : M2Nmatch,
            "message-id"       : M2Nonce|M2Nconv,
            "newsgroups"       : M2Nmust|M2Nrm,
            "nntp-posting-host": M2Nrm,
            "posted"           : M2Nrm,
            "posting-version"  : M2Nrm,
            "received"         : M2Nrm,
            "references"       : M2Nonce,
            "relay-version"    : M2Nrm,
            "sender"           : M2Nmatch,
            "subject"          : M2Nmust|M2Nonce,
            "to"               : M2Nmatch|M2Njoin,
            "mailing-list"     : M2Nmatch,
            "x-beenthere"      : M2Nmatch,
            "x-mailinglist"    : M2Nmatch,
            "x-mailing-list"   : M2Nmatch,
            "x-ml-name"        : M2Nmatch,
            "x-sequence"       : M2Nmatch,
            "x-uidl"           : M2Nrm,
            "x-ref"            : M2Nrm
            }

class Mail2NewsError(Exception):
    """Base class for all Mail2News exceptions"""
    def __init__(self, *args):
        apply(Exception.__init__, (self,)+args)
        try:
            self.error = args[0]
        except IndexError:
            self.error = "No error message given"

class M2NNoGroupMatched(Mail2NewsError):
    """No pattern matched, no group determined"""
    pass

class M2NNoMailRead(Mail2NewsError):
    """Tried to convert without email read before"""
    pass

class M2NMissingNewsArticle(Mail2NewsError):
    """Nothing converted to news article"""
    pass


class Mail2News:
    
    """Class to read mail, convert to news and post to a
    newsserver. readMail() will read an email message from stdin. This
    email can later be converted and posted to the server defined in
    the configuration. In case of failure the class tries to save the
    email, the converted news article if existent and a description of
    the error to files in a directory /tmp/m2n-XXXXXX-err/.

    The class the following public variables:

    confDict   - Dictionary containing the configuration
    
    mail       - File like object containing the email

    news       - File like object containing the converted news article
    
    message    - email.Message object
    
    newsgroups - List of the groups to which the converted article
                 will be posted. Initiated to [] in convert2News and filled
                 in checkTo().

    error      - A (debug|error) log
    """
    
    def __init__(self, server=None, port=119, user=None, passwd=None,
                 defgroup=None, pattern=[]):

        self.confDict = {"server"  : server,
                         "port"    : port,
                         "defgroup": defgroup,
                         "user"    : None,
                         "passwd"  : None,
                         "pattern" : pattern
                         }

        self.mail = None
        self.news = None
        self.Message = None
        self.newsgroups = []
        self.error = ""
        return

    def __deleteHeader(self, field):
        del self.message[field]

    def __joinHeader(self, field):
        values = self.message.get_all(field)
        if len(values) != 1:
            self.__deleteHeader(field)
            self.message[field] = Header(', '.join(values), None, None, field)

    def __deleteDuplicateHeader(self, field):
        if len(self.message.get_all(field)) != 1:
               value = self.message.get(field)
               self.__deleteHeader(field)
               self.message[field] = Header(value, None, None, field)
    
    # Call the appropriate method to convert a header with values to
    # news format.
    def __convertHeader(self, field):
        header = field.lower()
        if header == "message-id":
            self.__convertMID(field)
        elif header == "in-reply-to":
            self.__convertRef(field)
        
        # If we end here, I'm really stupid
        return "X-Cant-Convert: %s\nX-Cant-Convert: Report Bug to joerg@dietrich.net\n" % (header, )

    # Turn the email MID into a news MID, keep the old MID.
    def __convertMID(self, field):
        orgMID = self.message.get(field)
        newMID = self.__newsify(orgMID)
        if newMID != orgMID:
            self.message.replace_header(field, newMID)
            self.message['X-M2N-Message-ID'] = \
                                             Header(orgMID, None, None, 'X-M2N-Message-ID')


    # Turn In-Reply-To header to References.
    # Sometimes the In-Reply-To contains more information than just
    # the message-id
    # FIXME! Do something here about the really broken ones with only
    # one angle bracket. Check the checks for angle brackets in
    # __newsify()
    def __convertRef(self, field):
        orgREF = newREF = self.message.get('References', "")
        MIDs = self.message.get(field)
        self.__deleteHeader(field)
        for mid in string.split(MIDs, "\n"):
            newMID = self.__newsify(mid)
            if newMID not in newREF:
                newREF += "\n" + newMID
            if orgREF != newREF:
                if self.message.has_key('References'):
                    self.message.replace_header('References', newREF)
                else:
                    self.message['References'] = Header(newREF, None, None, 'References')
                    
                
    # Ensure that mail MIDs are in news format. Convert if necessary.
    # Some MUA generate MIDs with white spaces and/or no hostname
    # after the @
    def __newsify(self, MID):
        midSplit = string.split(MID.strip().replace(" ", ".").replace("<", "").replace(">", ""), "@")
        if len(midSplit) != 2:
            newMID = "<" + midSplit[0] + "@" + socket.getfqdn() + ">"
        # This is bogus, but we have to work around leafnode's
        # madness to reject localhost.valid.tld ...
        # So we have to perform the same checks as in leafnode's
        # validatefqdn.
        elif "." not in midSplit[1] \
                 or midSplit[1][:9] == "localhost"\
                 or midSplit[1][:8] == "127.0.0."\
                 or midSplit[1][:11] == "linux.local":
            newMID = "<" + midSplit[0] + "@" + socket.getfqdn() + ">"
        else:
            newMID =  "<" + midSplit[0] + "@" + midSplit[1] + ">"

        return newMID

    def __decodeHeader(self, str):
        decList = []
        for decoded, charset in email.Header.decode_header(str):
            if charset:
                decoded = unicode(decoded, charset)
            decList.append(decoded)
        return ' '.join(decList)
        
    #########################################################################
    ##                                                                     ##
    ## Here the "public" methods start.                                    ##
    ##                                                                     ##
    #########################################################################

    
    def saveError(self):

        """Try to save as much as possible in case of an error. Mail,
        news if applicable, and an error message will be saved to
        files mail, news, and error, respectively, in a directory
        /tmp/m2n-XXXXXX-err/. """
    
        saveDir = tempfile.mkdtemp("-err", "m2n-")
        if self.mail or self.message:
            if self.mail:
                saveFile = os.path.join(saveDir, "mail")
                f = open(saveFile, "w")
                self.mail.seek(0)
                f.write(self.mail.read())
                f.close()
            if self.message:
                saveFile = os.path.join(saveDir, "news")
                f = open(saveFile, "w")
                f.write(self.message.as_string())
                f.close()


        if not self.error:
            self.error += "Aborted without knowing the cause. Here is the "\
                          "most recent traceback:\n\n"
            tbTup = sys.exc_info()
            self.error += str(tbTup[0]) + "\n"
            self.error += str(tbTup[1]) + "\n"
            self.error += string.join(
                traceback.format_exception(None, None, tbTup[2]))
        saveFile = os.path.join(saveDir, "error")
        f = open(saveFile, "a")
        f.write(self.error)
        f.close()
        return

    def checkTo(self, value, debug=None):

        """Loops over all comma separated recipients in value and
        tries to match them with the regular expression patterns
        defined in the configuration. Matching groups are appended to
        the variable newsgroups, if not already present. """

        for pattern in self.confDict["pattern"]:
            regexp, group = pattern
            regExp = re.compile(unicode(regexp, 'utf-8'))
            for val in value:
                rcptList = string.split(val, ",")
                for rcpt in rcptList:
                    rcpt = string.strip(rcpt)
                    decodedSTR = self.__decodeHeader(rcpt)
                    if regExp.match(decodedSTR):
                        if debug:
                            print "\n-> %s matched %s\n" % (regexp, decodedSTR)
                        if group not in self.newsgroups:
                            self.newsgroups.append(group)
                    elif debug:
                        print "No match for %s and %s" % (regexp, decodedSTR)
        return


    def readMail(self):

        """Reads an email from stdin; makes a file like object stored
        in mail. """

        data = sys.stdin.read()
        self.mail = StringIO.StringIO(data)
        if self.mail:
            self.message = email.message_from_string(self.mail.getvalue())
        else:
            raise M2NNoMailRead("Tried to convert without email read before")
        return 

    def convert2News(self):

        """Convert the mail in \"mail\" to a news format article. The
        news article is stored in \"news\". newsgroups is initialized
        to [] and subsequently filled with target groups. """

        for field in self.message.keys():
            header = field.lower()
            if headDict.has_key(header):
                if headDict[header] & M2Nmatch:
                    self.checkTo(self.message.get_all(header))
                    
                if headDict[header] & M2Nrm:
                    self.__deleteHeader(field)
                    
                if headDict[header] & M2Njoin:
                    self.__joinHeader(field)

                if headDict[header] & M2Nrm:
                    self.__deleteHeader(field)
 
                if headDict[header] & M2Njoin:
                    self.__joinHeader(field)
 
                # If this header is only allowed once, we take only the
                # first instance; no matter how often it
                # appears. Robustness principle, yadda yadda yadda.
                if headDict[header] & M2Nonce:
                    self.__deleteDuplicateHeader(field)
 
                if headDict[header] & M2Nconv:
                    self.__convertHeader(field)
 
                # Unset M2Nmust for headers that already appeared
                # This is not ideal because it modifies headDict, which
                # then cannot be reused. We really should work on a copy here.
                if headDict[header] & M2Nmust and self.message.has_key(header):
                    headDict[header] ^= M2Nmust
                
        if len(self.newsgroups) == 0:
            if self.confDict["defgroup"]:
                self.newsgroups.append(self.confDict["defgroup"])
            else:
                self.error += "\nno match for newsgroups, no default group\n"
                raise M2NNoGroupMatched("no match for newsgroups, no default group")
            
        self.message['Newsgroups'] = Header(','.join(self.newsgroups), None, 
                                            None, 'Newsgroups')
        headDict["newsgroups"] ^= M2Nmust

        # Check to see if all mandatory headers are present.
        # If not, add the respective header without value and
        # hope that this is good enough. If not, the mail is
        # anyway to broken to be converted.
        for header in headDict.keys():
            if headDict[header] & M2Nmust:
                if header == "message-id":
                    self.message['Message-ID'] = \
                                               Header(email.Utils.make_msgid(), None, None, 'Message-ID')
                else:
                    self.message[header] = Header(None, None, None, header)
 
        self.message['X-M2N-Version'] = \
                                      Header(M2NVersion, None, None, 'X-M2N-Version')
        if not self.message.has_key('Approved'):
            self.message['Approved'] = \
                                     Header(pwd.getpwuid(os.getuid())[0] + "@" + socket.getfqdn(), None, None, 'Approved')
            
        self.news = StringIO.StringIO(self.message.as_string())
            
        return 0
                


    def postArt(self):

        """Post the article in self.news to the server defined in the
        configuration dictionary. If the connection does not already
        exist, it is initiated. The connection will be closed on
        deletion of the class. """

        if not self.news:
            raise M2NMissingNewsArticle("No news article present")
        
        self.news.seek(0)
        
        connection = nntplib.NNTP(self.confDict["server"],
                                       self.confDict["port"],
                                       self.confDict["user"],
                                       self.confDict["passwd"],
                                       1)
        
        reply = connection.post(self.news)

        connection.quit()
        return 0

    ###########################################################################
    #                                                                         #
    #                          END OF CLASS                                   #
    #                Mail2News end here, program starts below                 #
    #                                                                         #
    ###########################################################################
    

# Read the configuration file. If a configuration file is passed at
# instantiation, it will be the only file we attempt to read.
# Otherwise try "~/.mail2newsrc" and "/etc/mail2newsrc" in this order.
# If readinf the config file fails, we'll save the email to disk and
# raise an exception.
def readConfig(confFile=None):
    
    confList = [confFile, "~/.mail2newsrc", "/etc/mail2news.conf"]

    confDict = {"server"  : None,
                "port"    : 119,
                "defgroup": None,
                "user"    : None,
                "passwd"  : None,
                "pattern" : []
                }
    
    f = None
    for path in confList:
        if path == None:
            continue
        fullPath = os.path.expanduser(path)
        if checkConf(fullPath) == "ok":
            f = open(fullPath)
        else:
            if confList.index(path) == 0:
                break

    # Could not read config file
    if not f:
        confDict["server"] = -1
        return confDict
    
    for line in f.readlines():
        if line[0] == "#":
            continue
        if len(string.strip(line)) == 0:
            continue

        confLine = string.split(line, "=")
        if len(confLine) != 2:
            print "Illegal line in %s: %s" % (fullPath, confLine)
            continue
        opt, val = string.strip(confLine[0]), string.strip(confLine[1])

        if opt in confDict.keys():
            # Do not overwrite command line options
            # FIXME! implement command line options
            if not confDict[opt] or opt == "pattern":
                if opt == "port":
                    confDict[opt] = int(val)
                elif opt == "pattern":
                    confDict[opt].append(string.split(val))
                else:
                    confDict[opt] = val
        else:
            print "Unknown option: ", opt

    return confDict


# Check that the file in path exists and is a file
# FIXME! add permissions checks
def checkConf(path):
    
    path = os.path.realpath(path)
    if not os.path.exists(path):
        return "No such file or directory"
    if not os.path.isfile(path):
        return "Not a regular file"
    
    return "ok"



if __name__ == "__main__":

    # FIXME! Add command line options to override configuration file.
    # Maybe allow to fix a target group on the command line?
    confDict= readConfig()
    m2n = Mail2News(server=confDict["server"], port=confDict["port"],
                    user=confDict["user"], passwd=confDict["passwd"],
                    defgroup=confDict["defgroup"], pattern=confDict["pattern"])
    
    try:
        m2n.readMail()
        if confDict["server"] == -1:
            raise "Could not read config"
        if not m2n.convert2News():
            m2n.postArt()
    except:
        m2n.saveError()

